/*
 *  File Browse Utility, version 1.4
 *
 *  (c) 1992 Barry Nance
 *
 *  distributed through
 *  BYTE Magazine's Software Corner
*/

#include <stdio.h>
#include <dos.h>
#include <dir.h>
#include <bios.h>
#include <fcntl.h>
#include <conio.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <alloc.h>
#include <process.h>
#include <ctype.h>
#include <share.h>
#include <errno.h>

#define     BUFSIZE     4096
#define     BUFSIZEL    4096l

#define     STR_SIZE    1000

#define     TRUE           1
#define     FALSE          0

#define     BELL           7
#define     BS             8
#define     TAB            9
#define     LINEFEED      10
#define     FORMFEED      12
#define     CR            13
#define     BACKTAB       15
#define     ESC           27
#define     SPACE         32

#define     HOMEKEY       71
#define     ENDKEY        79
#define     UPKEY         72
#define     DOWNKEY       80
#define     PGUPKEY       73
#define     PGDNKEY       81
#define     LEFTKEY       75
#define     INSKEY        82
#define     RIGHTKEY      77
#define     DELKEY        83
#define     CTRLLEFTKEY  115
#define     CTRLRIGHTKEY 116
#define     F1            59
#define     F2            60
#define     F3            61
#define     F4            62
#define     F5            63
#define     F6            64
#define     F7            65
#define     F8            66
#define     F9            67
#define     F10           68

/************************************/

int     kbdstring(char buff[], int max_chars);
void    getkey(void);

int     fgetrecord(unsigned char buff[], int max_chars);
int     fgetline(unsigned char buff[], int max_chars);
int     fgetbyte(void);

int     fprevrecord(void);
int     fprevline(void);
int     fprevbyte(void);

void    drawbox(void);
void    gohome(void);
void    find_text(void);
void    showfile(void);
void    fill_window(void);
void    show_line_num(void);
void    display_line(void);
void    display_hex(void);
void    adjust_line_number(int amount);

int     get_video_mode(void);
void    save_cursor(struct csavetype *csave);
void    restore_cursor(struct csavetype *csave);
void    scroll_up(unsigned char numlines);
void    scroll_dn(unsigned char numlines);

int     search(unsigned char *pat, int plen, unsigned char *str, int slen);
void    search_setup(unsigned char *pat, int plen, int icase);

/************************************/

char     *fileprompt = "Filename? ";
char     *inv_arg    = "Invalid command line arguments.";
char     *notfound   = "File not found.";
char     *emptyfile  = "File is empty.";
char     *noaccess   = "File in use.  (Try later)";
char     *presskey   = "Press a key to continue.";

#define HELP_LINES 17
char    *help[HELP_LINES] =
            {
"                    File Browser 1.4       (c) 1992 Barry Nance           ",
"                                          BYTE Magazine's Software Corner ",
"                                                                          ",
"      Press:                                                              ",
"  ESC, X, or Q.....Quit                   W.........Toggle line wrap      ",
"  T or HOME........Top of file            H.........Toggle Hex/ASCII      ",
"  B or END.........Bottom of file         D.........Shell to DOS          ",
"  N................Next file              P.........Previous file         ",
"  PgUp or PgDn.....Up/Down a screenful                                    ",
"   or Spacebar..Next screenful                                         ",
"  \x1a.............Up/down/right/left                                     ",
"  F................Find text (any case)                                   ",
"  S................Search (exact case)                                    ",
"  A................Find/search again                                      ",
"                                                                          ",
"                                                                          ",
"                          <Press a key to resume>                         "
            };

unsigned char string[STR_SIZE];
char        string1[101];

char        search_arg[81];
int         search_flag;
int         ignore_case;
int         spos;

long        line_number;
long        linenum_save;

unsigned char CharCode;
unsigned char ScanCode;

int         xesc = FALSE;
int         eof_flag = FALSE;
int         start_of_file = TRUE;
int         line_end_char = 0;
int         hex_mode = FALSE;
int         wrap_mode = FALSE;

int         start_save;
int         eof_save;
int         fbufblock_save;
int         fbufndx_save;

unsigned    top, left, bottom, right, screen_width, screen_height;
int         left_relative_pos;

struct ftime ftimestruc;
struct ffblk findblock;

char        filename[80];
char        drive[5];
char        path[81];
char        name[10];
char        ext[5];

int         count;
int         flag;
int         file_num;
int         current_row;
int         file_is_empty;

int         screen_rcds;
int         screen_lines_hold;
int         screen_lines;
int         lines_read;
int         rcds_read;
int         lines_displayed;

unsigned    i, j, k;

int         fh;
long        filesize;
long        curr_filepos;
long        temp_long;
unsigned    num_blocks;
int         fbufndx;
int         fbufbytes;
unsigned    fbufblock;

char        *filelist_ptr;
char        *list_ptr;
char        *file_buff_ptr;
char        *screen_save_ptr;
char        *dos_screen_save_ptr;

char        line_save[160];

unsigned char low_inten;
unsigned char high_inten;
unsigned char found_attr;

struct      csavetype
            {
            unsigned int curloc;
            unsigned int curmode;
            };

struct  csavetype   cursor_data;
struct  csavetype   my_cpos;
struct  csavetype   dos_cpos;

union       REGS regs;

#define CSIZE   256             /* character set size */
static int skip[CSIZE];         /* increment to rightmost occurance of ch */
static int cmap[CSIZE];         /* upper to lowercase character map */


/************************************/

void    main(int argc, char *argv[])
        {
        if (get_video_mode() == 7)
            {
            low_inten   = 0x07;
            high_inten  = 0x0f;
            found_attr  = 0x70;
            }
        else
            {
            low_inten   = 0x02;
            high_inten  = 0x03;
            found_attr  = 0x0C;
            }

        file_buff_ptr = malloc(BUFSIZE+10);
        if (file_buff_ptr == NULL)
            {
            printf("Insufficient memory.\n");
            return;
            }
        memset(file_buff_ptr, 0, BUFSIZE+10);

        screen_save_ptr = malloc(4000);
        if (screen_save_ptr == NULL)
            {
            free(file_buff_ptr);
            printf("Insufficient memory.\n");
            return;
            }

        dos_screen_save_ptr = malloc(4000);
        if (dos_screen_save_ptr == NULL)
            {
            free(file_buff_ptr);
            free(screen_save_ptr);
            printf("Insufficient memory.\n");
            return;
            }

        filelist_ptr = calloc(1001, 13);
        if (filelist_ptr == NULL)
            {
            free(file_buff_ptr);
            free(screen_save_ptr);
            free(dos_screen_save_ptr);
            printf("Insufficient memory.\n");
            return;
            }
        list_ptr = filelist_ptr;
        count = 0;

        textattr(low_inten);
        gettext(1, 1, 80, 25, screen_save_ptr);
        save_cursor(&cursor_data);

        if (argc == 1)
            {
            window(1, 25, 80, 25);
            clrscr();
            window(1,  1, 80, 25);
            top = 1; left = 1; bottom = 24; right = 80;
            textattr(low_inten);
            gotoxy(1, 25);
            cputs(fileprompt);
            gotoxy(11, 25);
            kbdstring(filename, 70);
            goto p30;
            }

        if (argc == 2)
            {
            top = 1; left = 1; bottom = 24; right = 80;
            strcpy(filename, argv[1]);
            goto p30;
            }

        if (argc == 6)
            {
            top    = atoi(argv[2]);
            left   = atoi(argv[3]);
            bottom = atoi(argv[4]);
            right  = atoi(argv[5]);
            if (top == 0 || left == 0 || bottom == 0 || right == 0)
                goto p20;
            strcpy(filename, argv[1]);
            goto p30;
            }

p20:
        gotoxy(1, 24);
        textattr(high_inten);
        cputs(inv_arg);
        gohome();

p30:
        if (strlen(filename) == 0)
            gohome();

        if ( (flag = findfirst(filename, &findblock, 0)) != 0 )
            {
            printf("\n%s\n", notfound);
            free(file_buff_ptr);
            free(screen_save_ptr);
            free(dos_screen_save_ptr);
            free(filelist_ptr);
            exit(1);
            }

        screen_width  = right - left - 1;
        screen_height = bottom - top - 1;

        fnsplit(filename, drive, path, name, ext);

        while (flag == 0)
            {
            strcpy(list_ptr, findblock.ff_name);
            list_ptr += 13;
            count++;
            if (count >= 1000)
                break;
            flag = findnext(&findblock);
            }
        list_ptr  = filelist_ptr;
        file_num = 1;

open_a_file:
        drawbox();

        fnmerge(filename, drive, path, list_ptr, NULL);
        if (_osmajor < 2)
            fh = open(filename, O_RDONLY | O_BINARY);
        else
            fh = open(filename, O_RDONLY | O_BINARY | O_DENYNONE);

        if (fh == -1)
            {
            sound(800);
            delay(100);
            nosound();
            gotoxy(left+1, top+1);
            textattr(high_inten);
            if (errno == ENOENT)
                cputs(notfound);
            else
                cputs(noaccess);
            gohome();
            }

        getftime(fh, &ftimestruc);
        filesize   = filelength(fh);
        temp_long  = (filesize + (BUFSIZEL - 1l)) / BUFSIZEL;
        num_blocks = (unsigned) temp_long;

        start_of_file = TRUE;
        eof_flag  = FALSE;
        fbufblock = 0;
        fbufndx   = 0;
        lseek(fh, 0l, SEEK_SET);
        fbufbytes = read(fh, file_buff_ptr, BUFSIZE);

        if (fbufbytes == 0)
            file_is_empty = TRUE;
        else
            file_is_empty = FALSE;

        if (strchr(file_buff_ptr, LINEFEED) != NULL)
            line_end_char = LINEFEED;
        else
            line_end_char = CR;

        textattr(high_inten);
        window(1, 25, 80, 25);
        clrscr();
        window(1, 1, 80, 25);
        gotoxy(1, 25);
        cprintf(
"Line:       [            ]  %9ld bytes  %2d/%2d/%2d %2d:%2.2d:%2.2d  (F1=help)   ",
                      filesize,
                (int) ftimestruc.ft_month,
                (int) ftimestruc.ft_day,
                (int) ftimestruc.ft_year + 80,
                (int) ftimestruc.ft_hour,
                (int) ftimestruc.ft_min,
                (int) ftimestruc.ft_tsec);

        gotoxy(14, 25);
        textattr(high_inten);
        strcpy(string, list_ptr);
        while (strlen(string) < 12)
            strcat(string, " ");
        cputs(string);

        gotoxy(80, 25);

        CharCode = 0;
        ScanCode = HOMEKEY;
        showfile();

get_key:
        getkey();

        if (   CharCode == 'X' 
            || CharCode == 'x'
            || CharCode == 'Q'
            || CharCode == 'q'
            || CharCode == ESC)
                {
                close(fh);
                xesc = TRUE;
                gohome();
                }

        if (CharCode == 'T' || CharCode == 't')
            {
            CharCode = 0;
            ScanCode = HOMEKEY;
            }
        else
        if (CharCode == 'B' || CharCode == 'b')
            {
            CharCode = 0;
            ScanCode = ENDKEY;
            }

        if (CharCode == 'H' || CharCode == 'h')
            {
            lines_read  = 0;
            rcds_read   = 0;
            wrap_mode = FALSE;
            if (hex_mode)
                {
                hex_mode = FALSE;
                line_number = -1L;
                while (rcds_read < (bottom - top - 1) / 2)
                    if (fprevrecord() == -1)
                        break;
                }
            else
                {
                while (rcds_read < screen_rcds)
                    if (fprevrecord() == -1)
                        break;
                hex_mode = TRUE;
                }
            CharCode = 255;
            showfile();
            goto get_key;
            }

        if (CharCode == 'W' || CharCode == 'w')
            {
            if (hex_mode)
                goto get_key;
            rcds_read  = 0;
            lines_read = 0;
            while (rcds_read < screen_rcds)
                if (fprevrecord() == -1)
                    break;
            if (wrap_mode)
                wrap_mode = FALSE;
            else
                wrap_mode = TRUE;
            CharCode = 255;
            line_number = -1L;
            showfile();
            goto get_key;
            }

        if ( CharCode == '?' || (CharCode == 0 && ScanCode == F1) )
            {
            gettext(1, 1, 80, 25, dos_screen_save_ptr);
            save_cursor(&dos_cpos);
            textattr(low_inten);
            window(left+1, top+1, right-1, bottom-1);
            clrscr();
            window(1, 1, 80, 25);
            for (k=0; k<HELP_LINES; k++)
                {
                if (k+4 >= bottom) break;
                gotoxy(2, k+3);
                cputs(help[k]);
                }
            gotoxy(80, 25);
            getkey();
            puttext(1, 1, 80, 25, dos_screen_save_ptr);
            restore_cursor(&dos_cpos);
            goto get_key;
            }

        if (CharCode == 'D' || CharCode == 'd')
            {
            gettext(1, 1, 80, 25, dos_screen_save_ptr);
            save_cursor(&dos_cpos);
            puttext(1, 1, 80, 25, screen_save_ptr);
            restore_cursor(&cursor_data);
            textattr(high_inten);
            printf("\n");
            printf(
"         ͻ   \n");
            printf(
"                Type EXIT to resume file browsing.          \n");
            printf(
"         ͼ   \n");
            textattr(low_inten);
            system("");
            puttext(1, 1, 80, 25, dos_screen_save_ptr);
            restore_cursor(&dos_cpos);
            goto get_key;
            }

        if (CharCode == 'N' || CharCode == 'n')
            {
            if (file_num < count)
                {
                close(fh);
                list_ptr += 13;
                file_num++;
                goto open_a_file;
                }
            goto get_key;
            }

        if (CharCode == 'P' || CharCode == 'p')
            {
            if (file_num > 1)
                {
                close(fh);
                list_ptr -= 13;
                file_num--;
                goto open_a_file;
                }
            goto get_key;
            }

        if (CharCode == 0 || CharCode == 0xE0)
            if (   ScanCode == HOMEKEY
                || ScanCode == ENDKEY
                || ScanCode == PGUPKEY
                || ScanCode == PGDNKEY
                || ScanCode == UPKEY
                || ScanCode == DOWNKEY
                || ScanCode == LEFTKEY
                || ScanCode == RIGHTKEY)
                    {
                    showfile();
                    goto get_key;
                    }

        if (   CharCode == ' ' || CharCode == CR)
                {
                showfile();
                goto get_key;
                }

        if (CharCode == 'F' || CharCode == 'f'
            || CharCode == 'S' || CharCode == 's'
            || CharCode == 'A' || CharCode == 'a')
            {
            find_text();
            goto get_key;
            }

        goto get_key;
        }

/************************************/

void    drawbox(void)
        {
        int      i;

        textattr(low_inten);
        window(left, top, right, bottom);
        clrscr();
        window(1, 1, 80, 25);
        gotoxy(left, top);
        textattr(high_inten);
        cprintf("");

        gotoxy(left+1, top);
        cprintf("%.*s", (right - left) - 1,
"");

        gotoxy(right, top);
        cprintf("");

        for (i=top+1; i<bottom; i++)
            {
            gotoxy(left, i);
            cprintf("");
            gotoxy(right, i);
            cprintf("");
            }

        gotoxy(left, bottom);
        cprintf("");

        gotoxy(left+1, bottom);
        cprintf("%.*s", (right - left) - 1,
"");

        gotoxy(right, bottom);
        cprintf("");
        }

/************************************/

void    find_text(void)
        {
        gettext(1, 25, 80, 25, line_save);
        save_cursor(&my_cpos);
        window(1, 25, 80, 25);
        clrscr();
        window(1,  1, 80, 25);
        if ( strlen(search_arg) < 1 
          || (CharCode != 'A' && CharCode != 'a') )
            {
            textattr(high_inten);
            gotoxy(1, 25);
            if (CharCode == 'S' || CharCode == 's')
                {
                ignore_case = FALSE;
                cputs("Find(case-sen): ");
                gotoxy(17, 25);
                }
            else
                {
                ignore_case = TRUE;
                cputs("Find: ");
                gotoxy(7, 25);
                }
            kbdstring(search_arg, 50);
            if (strlen(search_arg) < 1)
                {
                puttext(1, 25, 80, 25, line_save);
                restore_cursor(&my_cpos);
                return;
                }
            }
        fbufblock_save = fbufblock;
        fbufndx_save   = fbufndx;
        start_save     = start_of_file;
        eof_save       = eof_flag;
        linenum_save   = line_number;
        search_setup(search_arg, strlen(search_arg), ignore_case);
        search_flag = 1;
        if (line_number == 1L
            && (CharCode != 'A' || CharCode != 'a'))
            {
            start_of_file = TRUE;
            eof_flag  = FALSE;
            fbufblock = 0;
            fbufndx   = 0;
            lseek(fh, 0l, SEEK_SET);
            fbufbytes = read(fh, file_buff_ptr, BUFSIZE);
            }
        lines_read = 0;
        gotoxy(1, 25);
        clreol();
        cputs("Looking... ");
        while (1)
            {
            if (fgetrecord(string, STR_SIZE) == -1)
                {
                gotoxy(1, 25);
                cputs("Not found.  <Press a key>");
                gotoxy(80, 25);
                getkey();
                break;
                }
            if ( (spos = search(search_arg, strlen(search_arg),
                          string, strlen(string))) != -1)
                {
                line_number = -1L;
                search_flag = 2;
                if (!wrap_mode && !hex_mode)
                    if (spos > left_relative_pos + screen_width)
                        {
                        left_relative_pos = spos - (screen_width / 2);
                        left_relative_pos = max(0, spos);
                        }
                    else
                        {
                        left_relative_pos = 0;
                        }
                break;
                }
            }
        puttext(1, 25, 80, 25, line_save);
        restore_cursor(&my_cpos);
        if (search_flag == 2)
            {
            showfile();
            }
        else
            {
            line_number   = linenum_save;
            start_of_file = start_save;
            eof_flag  = eof_save;
            fbufblock = fbufblock_save;
            fbufndx   = fbufndx_save;
            lseek(fh, ((long) fbufblock) * BUFSIZEL, SEEK_SET);
            fbufbytes = read(fh, file_buff_ptr, BUFSIZE);
            }
        search_flag = 0;
        }

/************************************/

void    adjust_line_number(int amount)
        {
        if (line_number == -1L)
            return;

        line_number += amount;
        if (line_number < 1L)
            line_number = 1L;
        }

/************************************/

void    showfile(void)
        {
        if (search_flag == 2)
            {
            rcds_read = 0;
            lines_read = 0;
            while (lines_read < screen_lines / 2)
                if (fprevline() == -1)
                    break;
            fill_window();
            show_line_num();
            return;
            }

        if (ScanCode == HOMEKEY)
            {
            left_relative_pos = 0;
            line_number = 1L;
            start_of_file = TRUE;
            eof_flag  = FALSE;
            fbufblock = 0;
            fbufndx   = 0;
            if (filesize == 0l)
                {
                sound(800);
                delay(100);
                nosound();
                gotoxy(left+1, top+1);
                textattr(high_inten);
                cputs(emptyfile);
                return;
                }
            else
                {
                lseek(fh, 0l, SEEK_SET);
                fbufbytes = read(fh, file_buff_ptr, BUFSIZE);
                fill_window();
                show_line_num();
                }
            return;
            }

        if (file_is_empty)
            return;

        if (ScanCode == PGUPKEY)
            {
            if (line_number == 1L)
                return;
            rcds_read  = 0;
            lines_read = 0;
            while (lines_read < screen_lines)
                if (fprevline() == -1)
                    {
                    lines_read = 0;
                    while (lines_read < screen_lines)
                        if (fgetline(string, STR_SIZE) == -1)
                            break;
                    return;
                    }
            lines_read = 0;
            rcds_read  = 0;
            while (lines_read < bottom - top - 1)
                if (fprevline() == -1)
                    {
                    line_number = 1L;
                    break;
                    }
            adjust_line_number(rcds_read * -1);
            fill_window();
            show_line_num();
            return;
            }

        if (ScanCode == PGDNKEY)
            {
            if (eof_flag)
                return;
            rcds_read = 0;
            screen_lines_hold = screen_lines;
            fill_window();
            adjust_line_number(screen_lines_hold);
            show_line_num();
            return;
            }

        if (ScanCode == UPKEY)
            {
            if (line_number == 1L)
                return;
            rcds_read  = 0;
            lines_read = 0;
            while (lines_read < screen_lines)
                if (fprevline() == -1)
                    {
                    lines_read  = 0;
                    while (lines_read < screen_lines)
                        if (fgetline(string, STR_SIZE) != -1)
                            break;
                    return;
                    }
            if (fprevline() != -1)
                {
                fgetline(string, STR_SIZE);
                scroll_dn(1);
                textattr(low_inten);
                current_row = top + 1;
                if (hex_mode)
                    display_hex();
                else
                    display_line();
                adjust_line_number(-1);
                }
            rcds_read  = 0;
            lines_read = 0;
            while (lines_read < screen_lines - 1)
                if (fgetline(string, STR_SIZE) == -1)
                    break;
            show_line_num();
            return;
            }

        if (ScanCode == DOWNKEY)
            {
            if (eof_flag)
                return;
            rcds_read  = 0;
            lines_read = 0;
            if (fgetline(string, STR_SIZE) != -1)
                {
                scroll_up(1);
                textattr(low_inten);
                current_row = bottom - 1;
                if (hex_mode)
                    display_hex();
                else
                    display_line();
                adjust_line_number(1);
                show_line_num();
                }
            return;
            }

        if (ScanCode == LEFTKEY)
            {
            if (left_relative_pos == 0)
                return;
            if (hex_mode || wrap_mode)
                return;
            movetext(left+1, top+1, right-6, bottom-1, 
                        left+6, top+1);
            left_relative_pos -= 5;
            rcds_read  = 0;
            lines_read = 0;
            while (lines_read < screen_lines)
                if (fprevline() == -1)
                    break;
            fill_window();
            show_line_num();
            return;
            }

        if (ScanCode == RIGHTKEY)
            {
            if (hex_mode || wrap_mode)
                return;
            if (left_relative_pos >= STR_SIZE - screen_width - 1)
                return;
            left_relative_pos += 5;
            rcds_read  = 0;
            lines_read = 0;
            while (lines_read < screen_lines)
                if (fprevline() == -1)
                    break;
            fill_window();
            show_line_num();
            return;
            }

        if (CharCode == CR || CharCode == 255 || CharCode == ' ')
            {
            if (eof_flag)
                return;
            if (CharCode == CR || CharCode == 255)
                left_relative_pos = 0;
            screen_lines_hold = screen_lines;
            rcds_read  = 0;
            lines_read = 0;
            fill_window();
            if (CharCode != 255)
                adjust_line_number(screen_lines_hold);
            show_line_num();
            return;
            }

        if (ScanCode == ENDKEY)
            {
            left_relative_pos = 0;
            line_number = -1L;
            start_of_file = FALSE;
            fbufblock = num_blocks - 1;
            lseek(fh, ((long) fbufblock) * BUFSIZEL, SEEK_SET);
            fbufbytes = read(fh, file_buff_ptr, BUFSIZE);
            fbufndx = fbufbytes - 1;
            rcds_read  = 0;
            lines_read = 0;
            while (lines_read < bottom - top - 2)
                if (fprevline() == -1)
                    break;
            fill_window();
            show_line_num();
            return;
            }
        }

/************************************/

void    fill_window(void)
        {
        textattr(low_inten);
        window(left+1, top+1, right-1, bottom-1);
        clrscr();
        window(1, 1, 80, 25);

        current_row = top + 1;
        screen_rcds = 0;
        screen_lines= 0;

        while (screen_lines < bottom - top - 1)
            {
            if (fgetrecord(string, STR_SIZE) == -1)
                break;
            screen_rcds++;
            textattr(low_inten);
            if ( search_flag == 2
              && (search(search_arg, strlen(search_arg),
                          string, strlen(string)) != -1) )
                textattr(found_attr);
            if (hex_mode)
                display_hex();
            else
                display_line();
            screen_lines += lines_displayed;
            }
        }

/************************************/

void    display_line(void)
        {
        int w, index;
        char s[81];

        w = 0;
        lines_displayed = 0;
        index = left_relative_pos;

        do  {
            window(left + 1, current_row, right - 1, current_row);
            clrscr();
            window(1, 1, 80, 25);
            gotoxy(left + 1, current_row);
            strncpy(s, &string[index], screen_width);
            s[screen_width] = '\0';
            cputs(s);
            current_row++;
            lines_displayed++;
            if (strlen(&string[index]) < screen_width + 1)
                return;
            index += screen_width;
            w += screen_width;
            }
        while (wrap_mode && current_row < bottom);
        }

/************************************/

void    show_line_num(void)
        {
        if (hex_mode || line_number == -1L)
            strcpy(string1, "     ");
        else
            sprintf(string1, "%5ld", line_number);

        textattr(high_inten);
        gotoxy(7, 25);
        cputs(string1);
        gotoxy(80, 25);
        }

/************************************/

void    display_hex(void)
        {
        int i, column_shift;
        unsigned char ch;

        column_shift = 10;
        if ( search_flag == 2
          && (search(search_arg, strlen(search_arg),
                      string, strlen(string)) != -1) )
            textattr(found_attr);
        else
            textattr(low_inten);

        gotoxy(left + 2, current_row);
        cprintf("%6.6lX ", curr_filepos);
        lines_displayed = 1;

        for (i=0; i<16; i++)
            {
            if (curr_filepos + i == filesize)
                break;
            ch = string[i];
            gotoxy(left + column_shift + (i * 3), current_row);
            cprintf("%2.2X ", (int) ch);
            if (i == 7)
                {
                cprintf("- ");
                column_shift = 12;
                }
            gotoxy(left + 61 + i, current_row);
            if (   ch == 0  || ch == BELL
                || ch == BS || ch == TAB
                || ch == LINEFEED || ch == CR  )
                ch = '.';
            cprintf("%c", ch);
            }

        current_row++;
        textattr(low_inten);
        }

/************************************/

int     fgetbyte(void)
        {
        eof_flag = FALSE;
        start_of_file = FALSE;
fgetbyte01:
        if (fbufblock == (num_blocks - 1) && fbufndx == fbufbytes)
            return -1;

        if (fbufndx < fbufbytes)
            return (int) file_buff_ptr[fbufndx++];

        fbufblock++;
        fbufndx = 0;
        lseek(fh, ((long) fbufblock) * BUFSIZEL, SEEK_SET);
        fbufbytes = read(fh, file_buff_ptr, BUFSIZE);
        goto fgetbyte01;
        }

/************************************/

int     fgetline(unsigned char buff[], int max_chars)
        {
        int  ch, i;

        if (hex_mode || !wrap_mode)
            return fgetrecord(buff, max_chars);

        memset(buff, 0, max_chars);
        i = 0;

        while (i < screen_width)
            {
            if ( (ch = fgetbyte()) == -1)
                {
                if (i == 0)
                    {
                    eof_flag = TRUE;
                    return -1;
                    }
                break;
                }
            if (ch == line_end_char || ch == 0)
                break;
            if (ch == CR || ch == LINEFEED || ch == BS)
                continue;
            if (ch == TAB)
                {
                do
                    buff[i++] = SPACE;
                while ( (i % 4) != 0 );
                continue;
                }
            if (ch == BELL)
                ch = SPACE;
            buff[i++] = (unsigned char) ch;
            }

        lines_read++;
        return 0;
        }

/************************************/

int     fgetrecord(unsigned char buff[], int max_chars)
        {
        int    i, c, rc;

        memset(buff, 0, max_chars);
        if (hex_mode)
            {
            curr_filepos = ( ((long) fbufblock) * BUFSIZEL) + fbufndx;
            for (i=0; i<16; i++)
                {
                if ( (c = fgetbyte()) == -1 )
                    if (i == 0 || search_flag != 0)
                        {
                        eof_flag = TRUE;
                        rc = -1;
                        goto fgs_exit;
                        }
                    else
                        {
                        lines_read++;
                        rc = 0;
                        goto fgs_exit;
                        }
                buff[i] = (unsigned char) c;
                }
            lines_read++;
            rc = 0;
            goto fgs_exit;
            }

        i = 0;
fgs1:   if (i == max_chars - 1)
            {
            if (wrap_mode)
                {
                c = (i + screen_width-1) / screen_width;
                if (c == 0) c = 1;
                lines_read += c;
                }
            else
                {
                lines_read++;
                }
            rc = 0;
            goto fgs_exit;
            }

        if ( (c = fgetbyte()) == -1 )
            {
            if (i == 0)
                {
                eof_flag = TRUE;
                rc = -1;
                goto fgs_exit;
                }
            if (wrap_mode)
                {
                c = (i + screen_width-1) / screen_width;
                if (c == 0) c = 1;
                lines_read += c;
                }
            else
                lines_read++;
            rc = 0;
            goto fgs_exit;
            }

        if (c == line_end_char || c == 0)
            {
            if (wrap_mode)
                {
                c = (i + screen_width-1) / screen_width;
                if (c == 0) c = 1;
                lines_read += c;
                }
            else
                lines_read++;
            rc = 0;
            goto fgs_exit;
            }

        if (c == CR || c == LINEFEED)
            goto fgs1;
        if (c == BS && i == 0)
            goto fgs1;

        if (c == TAB)
            {
            do
                buff[i++] = SPACE;
            while ( (i % 4) != 0 );
            goto fgs1;
            }

        if (c == BELL)
            c = SPACE;

        buff[i++] = (unsigned char) c;
        goto fgs1;


fgs_exit:
        if (rc != -1)
            rcds_read++;

        return rc;
        }

/************************************/

int     fprevbyte(void)
        {
        eof_flag = FALSE;
        start_of_file = FALSE;
sprev01:
        if (fbufblock == 0 && fbufndx <= 0)
            return -1;

        if (fbufndx > 0)
            return (int) file_buff_ptr[--fbufndx];

        fbufblock--;
        lseek(fh, ((long) fbufblock) * BUFSIZEL, SEEK_SET);
        fbufbytes = read(fh, file_buff_ptr, BUFSIZE);
        fbufndx = fbufbytes;
        goto sprev01;
        }

/************************************/

int     fprevline(void)
        {
        int  ch, i, j;

        if (hex_mode || !wrap_mode)
            return fprevrecord();

        if ( (ch = fprevbyte()) == -1)
            {
            start_of_file = TRUE;
            return -1;
            }

        i = 1;
        if (ch == line_end_char || ch == 0)
            ch = fprevbyte();

        while (ch != line_end_char && ch != 0 && ch != -1)
            {
            ch = fprevbyte();
            if (ch == CR || ch == LINEFEED || ch == BS)
                continue;
            i++;
            }

        j = 0;
        while (i > screen_width || j % (screen_width + 1) != 0)
            {
            ch = fgetbyte();
            i--;
            j++;
            }

        lines_read++;
        return 0;
        }

/************************************/

int     fprevrecord(void)
        {
        int     ch, slen, rc;

        if (hex_mode)
            {
            if (fprevbyte() == -1)
                {
                start_of_file = TRUE;
                rc = -1;
                goto g1l_exit;
                }
            while ( (fbufndx % 16) != 0 )
                {
                if (fprevbyte() == -1)
                    break;
                }
            lines_read++;
            rc = 0;
            goto g1l_exit;
            }

        if ( (ch = fprevbyte()) == -1)
            {
            start_of_file = TRUE;
            rc = -1;
            goto g1l_exit;
            }

        slen = 1;
        if (ch == line_end_char || ch == 0)
            ch = fprevbyte();

        while (ch != line_end_char && ch != 0 && ch != -1)
            {
            slen++;
            ch = fprevbyte();
            }

        if (ch != -1)
            {
            slen--;
            ch = fgetbyte();
            }

        if (wrap_mode)
            {
            ch = (slen + screen_width-1) / screen_width;
            if (ch == 0) ch = 1;
            lines_read += ch;
            }
        else
            lines_read++;

        rc = 0;

g1l_exit:
        if (rc != -1)
            rcds_read++;

        return rc;
        }

/************************************/

void    gohome(void)
        {
        if (xesc) goto restore_the_screen;

        textattr(high_inten);
        gotoxy(1, 25);
        cputs(presskey);
        getkey();

restore_the_screen:
        puttext(1, 1, 80, 25, screen_save_ptr);
        restore_cursor(&cursor_data);

        free(file_buff_ptr);
        free(screen_save_ptr);
        free(dos_screen_save_ptr);
        free(filelist_ptr);

        exit(0);
        }

/************************************/


void    save_cursor(struct csavetype *csave)
        {
        _AH = 3;
        _BH = 0;
        geninterrupt(0x10);
        csave->curloc  = _DX;
        csave->curmode = _CX;
        }


void    restore_cursor(struct csavetype *csave)
        {
        _DX = csave->curloc;
        _AH = 2;
        _BH = 0;
        geninterrupt(0x10);

        _CX = csave->curmode;
        _AH = 1;
        geninterrupt(0x10);
        }



int     kbdstring(char buff[], int max_chars)
        {
        int             ft, i, j, insert_mode, ctype, res;
        unsigned char   row, col, trow, tcol;
        unsigned int    cblock;

        ft = TRUE;
        i = j = insert_mode = 0;
        if (get_video_mode() == 7)
            cblock = 0x000D;
        else
            cblock = 0x0007;

        _AH = 3;
        _BH = 0;
        geninterrupt(0x10);
        ctype = _CX;
        col = wherex();
        row = wherey();

        cprintf("%-*s", max_chars-1, buff);
        gotoxy(col, row);

ks1:    getkey();
        tcol = wherex();
        trow = wherey();

        if (CharCode == ESC)
            {
            buff[0] = '\0';
            res = 0;
            goto kbdstring_exit;
            }

        if (CharCode == 0)
            {
            if (ScanCode == INSKEY)
                {
                if (insert_mode)
                    {
                    insert_mode = FALSE;
                    _CX = ctype;
                    _AH = 1;
                    geninterrupt(0x10);
                    }
                else
                    {
                    insert_mode = TRUE;
                    _CX = cblock;
                    _AH = 1;
                    geninterrupt(0x10);
                    }
                }
            else
            if (ScanCode == HOMEKEY)
                {
                i = 0;
                gotoxy(col, row);
                }
            else
            if (ScanCode == ENDKEY)
                {
                i = strlen(buff);
                gotoxy(col+strlen(buff), row);
                }
            else
            if (ScanCode == DELKEY)
                {
                for (j = i; j < strlen(buff); j++)
                    buff[j] = buff[j+1];
                gotoxy(col, row);
                cprintf("%-*s", max_chars-1, buff);
                gotoxy(tcol, trow);
                }
            else
            if (ScanCode == RIGHTKEY)
                {
                if (i < strlen(buff))
                    {
                    i++;
                    gotoxy(tcol+1, trow);
                    }
                }
            else
            if (ScanCode == LEFTKEY)
                {
                if (i > 0)
                    {
                    i--;
                    gotoxy(tcol-1, trow);
                    }
                }
            }

        if (CharCode == 0)
            goto ks1;

        if (CharCode == BS)
            {
            if (i > 0)
                {
                i--;
                gotoxy(tcol-1, trow);
                }
            }

        if (CharCode == CR)
            {
            res = 0;
            goto kbdstring_exit;
            }

        if (CharCode < 32)
            goto ks1;

        if (i == max_chars-1)
            goto ks1;

        if (ft)
            {
            ft = FALSE;
            if (i == 0 && !insert_mode)
                {
                memset(buff, 0, max_chars);
                gotoxy(col, row);
                cprintf("%-*s", max_chars-1, buff);
                gotoxy(col, row);
                }
            }

        if (insert_mode)
            {
            for (j = strlen(buff)-1; j >= i; j--)
                if (j < max_chars-2)
                    buff[j+1] = buff[j];
            buff[i++] = CharCode;
            _CX = ctype;
            _AH = 1;
            geninterrupt(0x10);
            gotoxy(col, row);
            cprintf("%-*s", max_chars-1, buff);
            gotoxy(++tcol, trow);
            _CX = cblock;
            _AH = 1;
            geninterrupt(0x10);
            }
        else
            {
            buff[i++] = CharCode;
            cprintf("%c", CharCode);
            }

        goto ks1;


kbdstring_exit:
        _CX = ctype;
        _AH = 1;
        geninterrupt(0x10);
        return res;
        }

void    getkey(void)
        {
        int     k;

        k        = bioskey(0);
        CharCode = k & 0x00FF;
        ScanCode = (k & 0xFF00) >> 8;
        }

int     get_video_mode(void)
        {
        regs.x.ax = 0x0f00;
        int86(0x10, &regs, &regs);
        regs.h.ah = 0;
        return regs.x.ax;
        }

void    scroll_up(unsigned char numlines)
        {
        regs.h.ah = 6;
        regs.h.al = numlines;
        regs.h.ch = (unsigned char) top;
        regs.h.cl = (unsigned char) left;
        regs.h.dh = (unsigned char) bottom-2;
        regs.h.dl = (unsigned char) right-2;
        regs.h.bh = (unsigned char) low_inten;
        int86(0x10, &regs, &regs);
        }

void    scroll_dn(unsigned char numlines)
        {
        regs.h.ah = 7;
        regs.h.al = numlines;
        regs.h.ch = (unsigned char) top;
        regs.h.cl = (unsigned char) left;
        regs.h.dh = (unsigned char) bottom-2;
        regs.h.dl = (unsigned char) right-2;
        regs.h.bh = (unsigned char) low_inten;
        int86(0x10, &regs, &regs);
        }

/************************************/

void    search_setup(pat,plen,icase)
                  unsigned char *pat;   /* pattern */
                  int plen;             /* pattern length */
                  int icase;            /* ignore case */

        {
        register unsigned char *p;
        register int i;

    /* initialize the skip array and character map */
        for (i = 0; i < CSIZE; ++i)
            {
            skip[i] = plen;
            cmap[i] = i;
            }

    /* map upper to lowercase on case insensitive searches */
        if (icase)
            for (i = 'A'; i <= 'Z'; ++i)
                cmap[i] = tolower(i);
    
    /* compute the skip values for characters in the pattern */
        for (i = plen-1, p = pat; i > 0; --i, ++p)
            skip[cmap[*p]] = i;
        }


int     search(pat,plen,str,slen)
                  unsigned char *pat;   /* pattern */
                  int plen;             /* pattern length */
                  unsigned char *str;   /* string */
                  int slen;             /* string length */
        {
        register unsigned char *pp,*pe,*sp,*se,*sb;

    /* exit immediately if the pattern is longer than the string */
        if (plen > slen)
            return -1;

    /* compute the pattern and string ends */
        pe = pat + plen;    
        se = str + slen;

    /* compute the starting point for the first match attempt */
        sb = str + plen - 1;

    /* match the pattern against the string */
        do {
            pp = pe - 1;
            sp = sb;
            while (pp >= pat && cmap[*sp] == cmap[*pp])
                {
                --sp;
                --pp;
                }
            if (pp < pat)
                return (sp - str + 1);
            }
        while ((sb += skip[cmap[*sb]]) < se);

    /* no match found */
        return -1;
        }


